/*
  WiFi connected round LED Clock. It gets NTP time from the internet and translates to a 60 RGB WS2812B LED strip.

  If you have another orientation where the wire comes out then change the methods getLEDHour and getLEDMinuteOrSecond

  Happy programming, Leon van den Beukel, march 2019

  ---  
  ------------------------------------------------------------------------------------------
  Brett
  Changed LED seconds to all on bar second which is Off. //Brett line 33 and line 122
  Board set to "LOLIN (WEMOS) D1 R2 amd MINI"
  v3 Added 30 sec sync and Adjust LED Brightness up or down every second when BriteUp D3 or BriteDown D4 are pressed
  v4 changed data pins 
  v6 all working syncs OK but v5 sync may work better
  v7  PIR
  v8/8a Mode switch to change display colours
  v9 added code to make hour and min LEDs the same as background colour
  v10 no change from v9
  v11 added extra print if PIR is OFF to check why brightness seting lost on PIR off sometimes
  

  Board Lolin WEMOS D1 R2 and Mini
  ------------------------------------------------------------------------------------------
  NTP and summer time code based on:
  https://tttapa.github.io/ESP8266/Chap15%20-%20NTP.html 
  https://github.com/SensorsIot/NTPtimeESP/blob/master/NTPtimeESP.cpp (for US summer time support check this link)  
  
*/

#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WiFiUdp.h>
#include <FastLED.h>
#define DEBUG_ON
const char ssid[] = "";// Your network SSID name here
const char pass[] = "";  // Your network password here
unsigned long timeZone = 0.0;                     // Change this value to your local timezone (in my case +1 for Amsterdam)
const char* NTPServerName = "uk.pool.ntp.org";    // Change this to a ntpserver nearby, check this site for a list of servers: https://www.pool.ntp.org/en/
unsigned long intervalNTP = 24 * 60 * 60000;      // Request a new NTP time every 24 hours

// Change the colors here if you want.
// Check for reference: https://github.com/FastLED/FastLED/wiki/Pixel-reference#predefined-colors-list
// You can also set the colors with RGB values, for example red:
// CRGB colorHour = CRGB(255, 0, 0);
CRGB colorHour = CRGB::Black;
CRGB colorMinute = CRGB::Black;
CRGB colorSecond = CRGB::Green;//Black;//Brett
CRGB colorHourMinute = CRGB::Black;
CRGB colorHourSecond = CRGB::Magenta;
CRGB colorMinuteSecond = CRGB::Magenta;
CRGB colorAll = CRGB::White;

// Set this to true if you want the hour LED to move between hours (if set to false the hour LED will only move every hour)
#define USE_LED_MOVE_BETWEEN_HOURS true

// Cutoff times for day / night brightness.
#define USE_NIGHTCUTOFF true   // Enable/Disable night brightness
#define MORNINGCUTOFF 23         // When does daybrightness begin?   8am
#define NIGHTCUTOFF 1          // When does nightbrightness begin? 10pm
int NIGHTBRIGHTNESS = 255;     // Brightness level from 0 (off) to 255 (full brightness)

ESP8266WiFiMulti wifiMulti;                     
WiFiUDP UDP;                                    
IPAddress timeServerIP;                         
const int NTP_PACKET_SIZE = 48;                 
byte NTPBuffer[NTP_PACKET_SIZE];                

unsigned long prevNTP = 0;
unsigned long lastNTPResponse = millis();
uint32_t timeUNIX = 0;
unsigned long prevActualTime = 0;

#define LEAP_YEAR(Y) ( ((1970+Y)>0) && !((1970+Y)%4) && ( ((1970+Y)%100) || !((1970+Y)%400) ) )
static const uint8_t monthDays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

#define BriteUp D3
#define BriteDown D4
int sync30 = 4; // 30 sec sync output
int sync30off = 1;//will go to 0 at 35 seconds to allow 30 sec pulse
int prev30secsync = 0;
int PIR = 5;// PIR input
int Mode = 15;// Display Mode switch
int ModeValue = 2; //Set display mode on startup (3 = Magenta with no other LEDs on- Black)
int BrightnesStore = 255;//stores LED brightness when PIR is off
//const char OffColour = 0; //the colour of the :LEDs when not on - eg Black off LED off, Yellow all off LEDs Yellow
#define NUM_LEDS 60     
#define DATA_PIN D6
CRGB LEDs[NUM_LEDS];

struct DateTime {
  int  year;
  byte month;
  byte day;
  byte hour;
  byte minute;
  byte second;
  byte dayofweek;
};

DateTime currentDateTime;

void setup() {
pinMode(BriteUp,INPUT_PULLUP); 
pinMode(BriteDown,INPUT_PULLUP);
pinMode(Mode,INPUT);
pinMode(PIR,INPUT);
pinMode(4 ,OUTPUT); 
  FastLED.delay(3000);
  FastLED.addLeds<WS2812B, DATA_PIN, GRB>(LEDs, NUM_LEDS);  

  Serial.begin(115200);          
  delay(10);
  Serial.println("\r\n");

  startWiFi();
  startUDP();

  if(!WiFi.hostByName(NTPServerName, timeServerIP)) { 
    Serial.println("DNS lookup failed. Rebooting.");
    Serial.flush();
    ESP.reset();
  }
  Serial.print("Time server IP:\t");
  Serial.println(timeServerIP);
  
  Serial.println("\r\nSending NTP request ...");
  sendNTPpacket(timeServerIP); 
 
}

void loop() {
  
  unsigned long currentMillis = millis();
 
  if (currentMillis - prevNTP > intervalNTP) { // If a minute has passed since last NTP request
    prevNTP = currentMillis;
    Serial.println("\r\nSending NTP request ...");
    sendNTPpacket(timeServerIP);               // Send an NTP request
  }

  uint32_t time = getTime();                   // Check if an NTP response has arrived and get the (UNIX) time
  if (time) {                                  // If a new timestamp has been received
    timeUNIX = time;
    Serial.print("NTP response:\t");
    Serial.println(timeUNIX);
    lastNTPResponse = currentMillis;
  } else if ((currentMillis - lastNTPResponse) > 3600000) {
    Serial.println("More than 1 hour since last NTP response. Rebooting.");
    Serial.flush();
    ESP.reset();
  }

  uint32_t actualTime = timeUNIX + (currentMillis - lastNTPResponse)/1000;
  if (actualTime != prevActualTime && timeUNIX != 0) { // If a second has passed since last update
     //#################################################################################################################
    //Adjust LED Brightness up or down every second when BriteUp D3 or BriteDown D4 are pressed
    brightness();
    //#################################################################################################################

   // #################################

    //Change display mode/LED colours
    setmode();

   // #################################

    prevActualTime = actualTime;
    convertTime(actualTime);
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    for (int i=0; i<NUM_LEDS; i++) 

    //Sets the colour of all the off LEDs
    
     if (ModeValue == 0)
 {
      LEDs[i] = CRGB::Magenta;//(242, 231, 48);//::Yellow;//Green;Black;//Brett


      // Set min and hour LEDs to match background colours if required.
      CRGB colorHour = CRGB::Black;
      CRGB colorMinute = CRGB::Black;

 }
   
    else if (ModeValue == 1)
 {
      LEDs[i] = CRGB::Yellow;//(242, 231, 48);//::Yellow;//Green;Black;//Brett
      // Set min and hour LEDs to match background colours if required.
     // CRGB colorHour = CRGB::Yellow;
    //  CRGB colorMinute = CRGB::Yellow;
 }

 
 else if (ModeValue == 2)
 {
      LEDs[i] = CRGB::Black;//(242, 231, 48);//::Yellow;//Green;Black;//Brett
// Set min and hour LEDs to match background colours if required.
      CRGB colorHour = CRGB::Black;
      CRGB colorMinute = CRGB::Black;
      CRGB colorHourSecond = CRGB::Magenta;
     CRGB colorMinuteSecond = CRGB::Magenta;
 }

else if (ModeValue == 3)
 {
      LEDs[i] = CRGB::Black;//(242, 231, 48);//::Yellow;//Green;Black;//Brett
// Set min and hour LEDs to match background colours if required.
      CRGB colorHour = CRGB::Black;
      CRGB colorMinute = CRGB::Black;
 }

 else if (ModeValue == 4)
 {
      LEDs[i] = CRGB::White;//(242, 231, 48);//::Yellow;//Green;Black;//Brett
// Set min and hour LEDs to match background colours if required.
      CRGB colorHour = CRGB::White;
      CRGB colorMinute = CRGB::White;
 }

   

    int second = getLEDMinuteOrSecond(currentDateTime.second);
    int minute = getLEDMinuteOrSecond(currentDateTime.minute);
    int hour = getLEDHour(currentDateTime.hour, currentDateTime.minute);

    // Set "Hands"
    LEDs[second] = colorSecond;

    //Disable hour and min colours so they are set by mode value.
   // LEDs[minute] = colorMinute;  
  //  LEDs[hour] = colorHour;  



    // Hour and min are on same spot
    if ( hour == minute)
      LEDs[hour] = colorHourMinute;

    // Hour and sec are on same spot
    if ( hour == second)
      LEDs[hour] = colorHourSecond;

    // Min and sec are on same spot
    if ( minute == second)
      LEDs[minute] = colorMinuteSecond;

    // All are on same spot
    if ( minute == second && minute == hour)
      LEDs[minute] = colorAll;

    if ( night() && USE_NIGHTCUTOFF == true )

    
      FastLED.setBrightness (NIGHTBRIGHTNESS); 
     
    FastLED.show();

    
    //################################################################################
  //30 second sync 
  
 
  
 
  if (second == 59 && sync30off == 0 )
  {
    digitalWrite(4, HIGH);
    Serial.println("30 second sync pulse ");
   
  }
    
    else 
    {
    digitalWrite(4, LOW);
    sync30off = 0;
   // Serial.println("sync30off = 0");
   // Serial.println("30 second sync pulse Off ");
  }
 //####################################################################################
  }  
}

byte getLEDHour(byte hours, byte minutes) {
  if (hours > 12)
    hours = hours - 12;

  byte hourLED;
  if (hours <= 5) 
    hourLED = (hours * 5) + 30;
  else
    hourLED = (hours * 5) - 30;

  if (USE_LED_MOVE_BETWEEN_HOURS == true) {
    if        (minutes >= 12 && minutes < 24) {
      hourLED += 1;
    } else if (minutes >= 24 && minutes < 36) {
      hourLED += 2;
    } else if (minutes >= 36 && minutes < 48) {
      hourLED += 3;
    } else if (minutes >= 48) {
      hourLED += 4;
    }
  }

  return hourLED;  
}

byte getLEDMinuteOrSecond(byte minuteOrSecond) {
  if (minuteOrSecond < 30) 
    return minuteOrSecond + 30;
  else 
    return minuteOrSecond - 30;
}

void startWiFi() { 
  wifiMulti.addAP(ssid, pass);   

  Serial.println("Connecting");
  byte i = 0;
  while (wifiMulti.run() != WL_CONNECTED) {  
    delay(250);
    Serial.print('.');
    LEDs[i++] = CRGB::Green;
    FastLED.show();    
  }
  Serial.println("\r\n");
  Serial.print("Connected to ");
  Serial.println(WiFi.SSID());             
  Serial.print("IP address:\t");
  Serial.print(WiFi.localIP());            
  Serial.println("\r\n");
}

void startUDP() {
  Serial.println("Starting UDP");
  UDP.begin(123);                          // Start listening for UDP messages on port 123
  Serial.print("Local port:\t");
  Serial.println(UDP.localPort());
  Serial.println();
}

uint32_t getTime() {
  if (UDP.parsePacket() == 0) { // If there's no response (yet)
    return 0;
  }
  UDP.read(NTPBuffer, NTP_PACKET_SIZE); // read the packet into the buffer
  // Combine the 4 timestamp bytes into one 32-bit number
  uint32_t NTPTime = (NTPBuffer[40] << 24) | (NTPBuffer[41] << 16) | (NTPBuffer[42] << 8) | NTPBuffer[43];
  // Convert NTP time to a UNIX timestamp:
  // Unix time starts on Jan 1 1970. That's 2208988800 seconds in NTP time:
  const uint32_t seventyYears = 2208988800UL;
  // subtract seventy years:
  uint32_t UNIXTime = NTPTime - seventyYears;
  return UNIXTime;
}

void sendNTPpacket(IPAddress& address) {
  memset(NTPBuffer, 0, NTP_PACKET_SIZE);  // set all bytes in the buffer to 0
  // Initialize values needed to form NTP request
  NTPBuffer[0] = 0b11100011;   // LI, Version, Mode
  // send a packet requesting a timestamp:
  UDP.beginPacket(address, 123); // NTP requests are to port 123
  UDP.write(NTPBuffer, NTP_PACKET_SIZE);
  UDP.endPacket();
}

void convertTime(uint32_t time) {
  // Correct time zone
  time += (3600 * timeZone);
  
  currentDateTime.second = time % 60;
  currentDateTime.minute = time / 60 % 60;
  currentDateTime.hour   = time / 3600 % 24;
  time  /= 60;  // To minutes
  time  /= 60;  // To hours
  time  /= 24;  // To days
  currentDateTime.dayofweek = ((time + 4) % 7) + 1;
  int year = 0;
  int days = 0;
  while ((unsigned)(days += (LEAP_YEAR(year) ? 366 : 365)) <= time) {
    year++;
  }
  days -= LEAP_YEAR(year) ? 366 : 365;
  time  -= days; // To days in this year, starting at 0  
  days = 0;
  byte month = 0;
  byte monthLength = 0;
  for (month = 0; month < 12; month++) {
    if (month == 1) { // February
      if (LEAP_YEAR(year)) {
        monthLength = 29;
      } else {
        monthLength = 28;
      }
    } else {
      monthLength = monthDays[month];
    }
  
    if (time >= monthLength) {
      time -= monthLength;
    } else {
      break;
    }
  }
 
  currentDateTime.day = time + 1;
  currentDateTime.year = year + 1970;
  currentDateTime.month = month + 1;  

  // Correct European Summer time
  if (summerTime()) {
    currentDateTime.hour += 1;
  }

#ifdef DEBUG_ON
  Serial.print(currentDateTime.year);
  Serial.print(" ");
  Serial.print(currentDateTime.month);
  Serial.print(" ");
  Serial.print(currentDateTime.day);
  Serial.print(" ");
  Serial.print(currentDateTime.hour);
  Serial.print(" ");
  Serial.print(currentDateTime.minute);
  Serial.print(" ");
  Serial.print(currentDateTime.second);
  Serial.print(" day of week: ");
  Serial.print(currentDateTime.dayofweek);
  Serial.print(" summer time: ");
  Serial.print(summerTime());
  Serial.print(" night time: ");
  Serial.print(night());  
  Serial.println();
#endif  
}

boolean summerTime() {

  if (currentDateTime.month < 3 || currentDateTime.month > 10) return false;  // No summer time in Jan, Feb, Nov, Dec
  if (currentDateTime.month > 3 && currentDateTime.month < 10) return true;   // Summer time in Apr, May, Jun, Jul, Aug, Sep
  if (currentDateTime.month == 3 && (currentDateTime.hour + 24 * currentDateTime.day) >= (3 +  24 * (31 - (5 * currentDateTime.year / 4 + 4) % 7)) || currentDateTime.month == 10 && (currentDateTime.hour + 24 * currentDateTime.day) < (3 +  24 * (31 - (5 * currentDateTime.year / 4 + 1) % 7)))
  return true;
    else
  return false;
}

boolean night() {
  
  if (currentDateTime.hour >= NIGHTCUTOFF && currentDateTime.hour <= MORNINGCUTOFF) 
    return true;    
}


//#################################
//Adjust brighness
void brightness() {
  if(digitalRead(PIR) == LOW && NIGHTBRIGHTNESS >0)
  {
   BrightnesStore = NIGHTBRIGHTNESS;// saves brightness level for when PIR comes on
   Serial.println("");
  Serial.print("BrightnesStore1 ");  
  Serial.println(BrightnesStore);
  NIGHTBRIGHTNESS = 0;
    }
    else if (digitalRead(PIR) == HIGH )
    {
  NIGHTBRIGHTNESS = BrightnesStore;
  if(digitalRead(BriteUp) == LOW && NIGHTBRIGHTNESS < 255)
  {
     

    NIGHTBRIGHTNESS = NIGHTBRIGHTNESS + 25;
     Serial.print("NIGHTBRIGHTNESS UP ");
     Serial.println(NIGHTBRIGHTNESS);
  }

    if(digitalRead(BriteDown) == LOW && NIGHTBRIGHTNESS > 0)
   {
    NIGHTBRIGHTNESS = NIGHTBRIGHTNESS - 25;
    Serial.print("NIGHTBRIGHTNESS DOWN ");
    Serial.println(NIGHTBRIGHTNESS);
  }
  // sets min max range of LED brightness
if ( NIGHTBRIGHTNESS > 255 )
{
  NIGHTBRIGHTNESS = 255;
  Serial.print("NIGHTBRIGHTNESS MAX LIMIT REACHED ");
    Serial.println(NIGHTBRIGHTNESS);
}
  else if ( NIGHTBRIGHTNESS < 0 )
{
  NIGHTBRIGHTNESS = 0;
  Serial.print("NIGHTBRIGHTNESS MIN LIMIT REACHED ");
    Serial.println(NIGHTBRIGHTNESS);
}
BrightnesStore = NIGHTBRIGHTNESS;
Serial.println("");
Serial.print("BrightnesStore2 ");  
  Serial.println(BrightnesStore);


    }

    if(digitalRead(PIR) == LOW)
    {
      Serial.println("");
Serial.print("PIR OFF ");  
  Serial.println(BrightnesStore);
    }
    
}


//############################################
void setmode(){
 //ModeValue = digitalRead(Mode);
if (digitalRead(Mode) == HIGH)
{
  ModeValue = ModeValue +1;
  colorSecond = CRGB::Blue;//Black;//Brett
  Serial.println("Secs Colour Blue ");

  if(ModeValue == 5)// Sets the number of modes eg 5 will be 4 modes add modes to the if statements below
  {
    ModeValue = 0;
  }
    
}
 // Set colour of Seconds LED. ## Note background LED colour set on line 168 
  if (ModeValue == 0)
 {

   colorSecond = CRGB::Blue;
   
   Serial.print("Secs Colour Blue/Yellow Mode Value ");
    Serial.println(ModeValue);
}

 
 if (ModeValue == 1)
 {

   colorSecond = CRGB::Red;
   
   Serial.print("Secs Colour Red/Yellow Mode Value ");
    Serial.println(ModeValue);
}

if (ModeValue == 2)
 {
colorSecond = CRGB::Magenta;
CRGB colorHourSecond = CRGB::Magenta;
CRGB colorMinuteSecond = CRGB::Magenta;
 
  Serial.print("Secs Colour Magenta/Black Mode Value ");
  Serial.println(ModeValue);
 }

if (ModeValue == 3)
 {
colorSecond = CRGB::White;
 
  Serial.print("Secs Colour White/Black Mode Value ");
  Serial.println(ModeValue);
 }


if (ModeValue == 4)
 {
colorSecond = CRGB::Red;
 
  Serial.print("Secs Colour Red/White Mode Value ");
  Serial.println(ModeValue);
 }

 


 
}
